#include <assert.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>

#include "gpio.h"
#define BUF_SIZE	3
#define DIRECTION_MAX	35
#define VALUE_MAX	30

static int gpio_export(int pin);
static int gpio_unexport(int pin);
static int gpio_set_direction(int pin, int dir);


int gpio_open(const struct driver *drv)
{
    int res;

    assert(NULL != drv);

    struct gpio *this = (struct gpio *)(drv->dev);

    res = gpio_unexport(this->pin);
    res = gpio_export(this->pin);
    usleep(100000); /* Wait until sysfs is exported. */

    return res;
}

int gpio_close(const struct driver *drv)
{
    int res;

    assert(NULL != drv);

    struct gpio *this = (struct gpio *)(drv->dev);

    res = gpio_unexport(this->pin);

    return res;
}

int gpio_read(const struct driver *drv, char *buffer, unsigned int len)
{
    assert(NULL != drv);

    char path[VALUE_MAX];
    int fd;

    struct gpio *this = (struct gpio *)(drv->dev);

    snprintf(path, VALUE_MAX, "/sys/class/gpio/gpio%d/value", this->pin);
    fd = open(path, O_RDONLY);
    if(-1 == fd) {
        return -1;
    }

    if(-1 == read(fd, buffer, len)) {
        return -1;
    }
    close(fd);

    return 0;
}

int gpio_write(const struct driver *drv, const char *buffer, unsigned int len)
{
    assert(NULL != drv);

    char path[VALUE_MAX];
    int fd;

    struct gpio *this = (struct gpio *)(drv->dev);

    snprintf(path, VALUE_MAX, "/sys/class/gpio/gpio%d/value", this->pin);
    fd = open(path, O_WRONLY);
    if(-1 == fd) {
        return -1;
    }

    int res;
    if((res = write(fd, buffer, len)) != 1) {
        return -1;
    }

    close(fd);


    return 0;
}

int gpio_ioctl(const struct driver *drv, unsigned int cmd, unsigned int argc, va_list args)
{
    assert(drv != 0);

    int res = 0, dir;
    struct gpio *this = (struct gpio *)(drv->dev);

    switch(cmd) {
        case IOCTL_CMD_SET_DIRECTION:
            if(argc != 1) {
                res = -1;
                break;
            }
            dir = va_arg(args, int);
            res = gpio_set_direction(this->pin, dir);
            break;
    }
    return res;
}

static int gpio_export(int pin)
{
    char buffer[BUF_SIZE];
    ssize_t bytes_written;
    int fd;

    fd = open("/sys/class/gpio/export", O_WRONLY);
    if(-1 == fd) {
        return -1;
    }

    bytes_written = snprintf(buffer, BUF_SIZE, "%d", pin);
    write(fd, buffer, bytes_written);
    close(fd);

    return 0;
}

static int gpio_unexport(int pin)
{
    char buffer[BUF_SIZE];
    ssize_t bytes_written;
    int fd;

    fd = open("/sys/class/gpio/unexport", O_WRONLY);
    if(-1 == fd) {
        return -1;
    }

    bytes_written = snprintf(buffer, BUF_SIZE, "%d", pin);
    write(fd, buffer, bytes_written);
    close(fd);

    return 0;
}

static int gpio_set_direction(int pin, int dir)
{
    static const char directions_str[]  = "in\0out";
    char path[DIRECTION_MAX];
    int fd;

    snprintf(path, DIRECTION_MAX, "/sys/class/gpio/gpio%d/direction", pin);
    fd = open(path, O_WRONLY);
    if(-1 == fd) {
        return -1;
    }

    if(-1 == write(fd, &directions_str[0 == dir ? 0 : 3], 0 == dir ? 2 : 3)) {
        return -1;
    }
    close(fd);

    return 0;
}
